#!/usr/bin/env python

class CSV_Parser(object):
    """
    Examples
    --------
    >>> from pprint import pprint

    >>> from jsonextended.utils import MockPath
    >>> fileobj = MockPath(is_file=True,
    ... content='''# comment line
    ... head1,head2
    ... val1,val2
    ... val3,val4'''
    ... )
    >>> with fileobj.open() as f:
    ...     data = CSV_Parser().read_file(f)
    >>> pprint(data)
    {'head1': ['val1', 'val3'], 'head2': ['val2', 'val4']}

    """

    plugin_name = 'csv.basic'
    plugin_descript = 'read *.csv delimited file with headers to {header:[column_values]}'
    file_regex = '*.csv'

    def read_file(self, file_obj, **kwargs):

        delim = kwargs.get('csv_delim', ',')
        comments = kwargs.get('comments', '#')
        keypairs = None
        for line in file_obj:
            if line.strip().startswith(comments):
                continue
            values = line.strip().split(delim)
            if keypairs is None:
                keypairs = [(v, []) for v in values]
                continue
            assert len(keypairs) == len(values), 'row different length to headers'
            for keypair, value in zip(keypairs, values):
                keypair[1].append(value)

        return {} if keypairs is None else dict(keypairs)



from pint import UnitRegistry

ureg = UnitRegistry()
from pint.quantity import _Quantity


class Encode_Quantity(object):
    """

    Examples
    --------
    >>> from pprint import pprint
    >>> from pint import UnitRegistry
    >>> ureg = UnitRegistry()

    >>> print(Encode_Quantity().to_str(ureg.Quantity(1,'nanometre')))
    1 nm

    >>> pprint(Encode_Quantity().to_json(ureg.Quantity(1,'nanometre')))
    {'_pint_Quantity_': {'Magnitude': 1, 'Units': 'nanometer'}}

    >>> Encode_Quantity().from_json({'_pint_Quantity_': {'Magnitude': 1, 'Units': 'nanometer'}})
    <Quantity(1, 'nanometer')>

    """

    plugin_name = 'pint.Quantity'
    plugin_descript = 'encode/decode pint.Quantity object'
    objclass = _Quantity
    dict_signature = ['_pint_Quantity_']

    def to_str(self, obj):
        return ' '.join(u'{:~}'.format(obj).split())

    def to_json(self, obj):
        value = obj.magnitude
        units = obj.units
        return {'_pint_Quantity_': {'Magnitude': value, 'Units': str(units)}}

    def from_json(self, obj):
        return ureg.Quantity(obj['_pint_Quantity_']['Magnitude'],
                             obj['_pint_Quantity_']['Units'])



#!/usr/bin/env python
import ast


class CSVLiteral_Parser(object):
    """
    Examples
    --------
    >>> from pprint import pprint

    >>> from jsonextended.utils import MockPath
    >>> fileobj = MockPath(is_file=True,
    ... content='''# comment line
    ... head1,head2
    ... 1.1,3
    ... 2.2,"3.3"'''
    ... )
    >>> with fileobj.open() as f:
    ...     data = CSVLiteral_Parser().read_file(f)
    >>> pprint(data)
    {'head1': [1.1, 2.2], 'head2': [3, '3.3']}

    """

    plugin_name = 'csv.literal'
    plugin_descript = "read *.literal.csv delimited files with headers to {header:column_values}, with number strings converted to int/float"
    ", s.t. values are converted to their python type"
    file_regex = '*.literal.csv'

    @staticmethod
    def tryeval(val):
        try:
            val = ast.literal_eval(val)
        except ValueError:
            pass
        return val

    def read_file(self, file_obj, **kwargs):

        delim = kwargs.get('csv_delim', ',')
        comments = kwargs.get('comments', '#')
        keypairs = None
        for line in file_obj:
            if line.strip().startswith(comments):
                continue
            values = line.strip().split(delim)
            if keypairs is None:
                keypairs = [(v, []) for v in values]
                continue
            assert len(keypairs) == len(values), 'row different length to headers'
            for keypair, value in zip(keypairs, values):
                keypair[1].append(value)

        return {} if keypairs is None else {k: [self.tryeval(v) for v in vs] for k, vs in keypairs}



#!/usr/bin/env python
"""mock files and folder structure for testing

Examples
--------
>>> jsonfile1
MockFile("dir1/file1.json")
>>> jsonfile2
MockFile("file2.json")
>>> csvfile1
MockFile("dir1/subdir1/file1.csv")
>>> csvfile2
MockFile("dir1/subdir1/file1.literal.csv")
>>> kpfile1
MockFile("dir1/subdir2/subsubdir21/file1.keypair")

>>> print(directory1.to_string(indentlvl=3,file_content=False))
Folder("dir1")
   File("file1.json")
   Folder("subdir1")
     File("file1.csv")
     File("file1.literal.csv")
   Folder("subdir2")
     Folder("subsubdir21")
       File("file1.keypair")

>>> print(directory1.to_string(indentlvl=3,file_content=True))
Folder("dir1")
   File("file1.json") Contents:
    {"key2": {"key3": 4, "key4": 5}, "key1": [1, 2, 3]}
   Folder("subdir1")
     File("file1.csv") Contents:
       # a csv file
      header1,header2,header3
      val1,val2,val3
      val4,val5,val6
      val7,val8,val9
     File("file1.literal.csv") Contents:
       # a csv file with numbers
      header1,header2,header3
      1,1.1,string1
      2,2.2,string2
      3,3.3,string3
   Folder("subdir2")
     Folder("subsubdir21")
       File("file1.keypair") Contents:
         # a key-pair file
        key1 val1
        key2 val2
        key3 val3
        key4 val4

"""

from jsonextended.utils import MockPath

jsonfile1 = MockPath('file1.json', is_file=True,
                     content='{"key2": {"key3": 4, "key4": 5}, "key1": [1, 2, 3]}')

jsonfile2 = MockPath('file2.json', is_file=True,
                     content='{"key1":{"_python_set_": [1, 2, 3]},"key2":{"_numpy_ndarray_": {"dtype": "int64", "value": [1, 2, 3]}}}')

csvfile1 = MockPath('file1.csv', is_file=True,
                    content=""" # a csv file
header1,header2,header3
val1,val2,val3
val4,val5,val6
val7,val8,val9
""")

csvfile2 = MockPath('file1.literal.csv', is_file=True,
                    content=""" # a csv file with numbers
header1,header2,header3
1,1.1,string1
2,2.2,string2
3,3.3,string3
""")

kpfile1 = MockPath('file1.keypair', is_file=True,
                   content=""" # a key-pair file
key1 val1
key2 val2
key3 val3
key4 val4
""")

directory1 = MockPath('dir1',
                      structure=[jsonfile1, {'subdir1': [csvfile1, csvfile2]},
                                 {'subdir2': [{'subsubdir21': [kpfile1]}]}]
                      )



import contextlib
import os
import sys
import tempfile
from fnmatch import fnmatch
from functools import total_ordering
import collections

# python 2/3 compatibility
try:
    basestring
except NameError:
    basestring = str

try:
    import pathlib
except ImportError:
    import pathlib2 as pathlib

try:
    unicode
except NameError:
    unicode = str


# TODO handling bytes/encoding (py 2 and 3)
class _OpenRead(object):
    def __init__(self, linelist, encoding=None, usebytes=False):
        self._linelist = linelist
        self._current_indx = 0
        self._current_line = 0
        self._encoding = encoding
        self._bytes = usebytes

    def read(self, size=None):
        out = '\n'.join(self._linelist)[self._current_indx:]
        if size is not None:
            self._current_indx += size
            out = out[0:self._current_indx]
        if self._encoding is not None:
            out = out.encode(self._encoding)
        return out

    def readline(self):
        if self._current_line >= len(self._linelist):
            line = ''
        else:
            line = self._linelist[self._current_line] + '\n'
        self._current_line += 1
        self._current_indx += len('\n'.join(self._linelist[0:self._current_line]))
        if self._encoding is not None:
            line = line.encode(self._encoding)
        return line

    def readlines(self):
        self._current_indx = len('\n'.join(self._linelist))
        self._current_line = len(self._linelist)
        if self._encoding is not None and self:
            return [line.encode(self._encoding) for line in self._linelist]
        else:
            return self._linelist[:]

    def __iter__(self):
        for line in self._linelist:
            if self._encoding is not None:
                line = line.encode(self._encoding)
            yield line


# TODO handling bytes/encoding (py 2 and 3)
class _OpenWrite(object):
    def __init__(self, usebytes=False):
        self._str = ''
        self._bytes = usebytes

    def write(self, instr):
        if hasattr(instr, "decode"):
            instr = instr.decode()
        self._str += instr

    def writelines(self, lines):
        for instr in lines:
            if hasattr(instr, "decode"):
                instr = instr.decode()
            self.write(instr)


@total_ordering
class MockPath(object):
    r"""a mock path, mimicking pathlib.Path,
    supporting context open method for read/write

    Parameters
    ----------
    path : str
        the path string
    is_file : bool
        if True is file, else folder
    content : str
        content of the file
    structure:
        structure of the directory

    Examples
    --------
    >>> file_obj = MockPath("path/to/test.txt",is_file=True,
    ...                     content="line1\nline2\nline3")
    ...
    >>> file_obj
    MockFile("path/to/test.txt")
    >>> file_obj.name
    'test.txt'
    >>> file_obj.parent
    MockFolder("path/to")
    >>> print(str(file_obj))
    path/to/test.txt
    >>> print(file_obj.to_string())
    File("test.txt") Contents:
    line1
    line2
    line3
    >>> file_obj.is_file()
    True
    >>> file_obj.is_dir()
    False
    >>> with file_obj.open('r') as f:
    ...     print(f.readline().strip())
    line1
    >>> with file_obj.open('w') as f:
    ...     f.write('newline1\nnewline2')
    >>> print(file_obj.to_string())
    File("test.txt") Contents:
    newline1
    newline2

    >>> with file_obj.maketemp() as temp:
    ...     with temp.open() as f:
    ...         print(f.readline().strip())
    newline1

    >>> dir_obj = MockPath(
    ...   structure=[{'dir1':[{'subdir':[file_obj.copy_path_obj()]},file_obj.copy_path_obj()]},
    ...              {'dir2':[file_obj.copy_path_obj()]},file_obj.copy_path_obj()]
    ... )
    >>> dir_obj
    MockFolder("root")
    >>> dir_obj.name
    'root'
    >>> dir_obj.is_file()
    False
    >>> dir_obj.is_dir()
    True
    >>> print(dir_obj.to_string())
    Folder("root")
      Folder("dir1")
        Folder("subdir")
          File("test.txt")
        File("test.txt")
      Folder("dir2")
        File("test.txt")
      File("test.txt")

    >>> "dir1/test.txt" in dir_obj
    True

    >>> dir_obj["dir1/test.txt"]
    MockFile("root/dir1/test.txt")

    >>> list(dir_obj.iterdir())
    [MockFolder("root/dir1"), MockFolder("root/dir2"), MockFile("root/test.txt")]

    >>> list(dir_obj.glob("*"))
    [MockFolder("root/dir1"), MockFolder("root/dir2"), MockFile("root/test.txt")]

    >>> list(dir_obj.glob("dir1/*"))
    [MockFolder("root/dir1/subdir"), MockFile("root/dir1/test.txt")]

    >>> list(dir_obj.glob("**"))
    [MockFolder("root/dir1"), MockFolder("root/dir1/subdir"), MockFolder("root/dir2")]

    >>> list(dir_obj.glob("**/*"))
    [MockFolder("root/dir1"), MockFolder("root/dir1/subdir"), MockFile("root/dir1/subdir/test.txt"), MockFile("root/dir1/test.txt"), MockFolder("root/dir2"), MockFile("root/dir2/test.txt"), MockFile("root/test.txt")]

    #>>> list(dir_obj.glob("**/dir1"))
    #[MockFolder("root/dir1")]

    >>> new = dir_obj.joinpath('dir3')
    >>> new.mkdir()
    >>> list(dir_obj.iterdir())
    [MockFolder("root/dir1"), MockFolder("root/dir2"), MockFolder("root/dir3"), MockFile("root/test.txt")]

    >>> dir_obj.joinpath("test.txt").unlink()
    >>> list(dir_obj.iterdir())
    [MockFolder("root/dir1"), MockFolder("root/dir2"), MockFolder("root/dir3")]

    >>> dir_obj.joinpath("dir3").rmdir()
    >>> list(dir_obj.iterdir())
    [MockFolder("root/dir1"), MockFolder("root/dir2")]

    >>> print(dir_obj.to_string())
    Folder("root")
      Folder("dir1")
        Folder("subdir")
          File("test.txt")
        File("test.txt")
      Folder("dir2")
        File("test.txt")

    >>> dir_obj.joinpath("dir1/subdir")
    MockFolder("root/dir1/subdir")

    >>> dir_obj.joinpath("dir1", "subdir")
    MockFolder("root/dir1/subdir")

    >>> new = dir_obj.joinpath("dir1/subdir/other")
    >>> new
    MockVirtualPath("root/dir1/subdir/other")

    >>> new.touch()
    >>> new
    MockFile("root/dir1/subdir/other")

    >>> new.unlink()
    >>> new
    MockVirtualPath("root/dir1/subdir/other")

    >>> new.mkdir()
    >>> new
    MockFolder("root/dir1/subdir/other")

    >>> newfile = MockPath('newfile.txt', is_file=True)
    >>> new.copy_from(newfile)
    >>> print(new.to_string())
    Folder("other")
      File("newfile.txt")

    >>> file_obj = MockPath("newfile2.txt",is_file=True, content='test')
    >>> file_obj.copy_to(new)
    >>> print(new.to_string())
    Folder("other")
      File("newfile.txt")
      File("newfile2.txt")

    >>> file_obj.name = "newfile3.txt"
    >>> with file_obj.maketemp() as temp:
    ...    new.copy_from(temp)
    >>> print(new.to_string())
    Folder("other")
      File("newfile.txt")
      File("newfile2.txt")
      File("newfile3.txt")

    >>> print(new.copy_path_obj().to_string())
    Folder("other")
      File("newfile.txt")
      File("newfile2.txt")
      File("newfile3.txt")

    >>> with new.maketemp(getoutput=True) as tempdir:
    ...     tempdir.joinpath("new").mkdir()
    ...     tempdir.joinpath("new/file.txt").touch()
    >>> print(new.to_string())
    Folder("other")
      Folder("new")
        File("file.txt")
      File("newfile.txt")
      File("newfile2.txt")
      File("newfile3.txt")

    """

    def __init__(self, path='root',
                 is_file=False, exists=True,
                 structure=(), content='', parent=None):
        self._path = path
        self._name = os.path.basename(path)
        self._exists = exists
        self._is_file = is_file
        self._is_dir = not is_file
        self._content = content.splitlines()
        self._parent = parent
        self._children = []

        paths = self._splitall(path)
        if len(paths) > 1 and parent is None:
            self.parent = MockPath(os.path.join(*paths[:-1]))

        for subobj in structure:
            if hasattr(subobj, 'keys'):
                key = list(subobj.keys())[0]
                self.add_child(MockPath(os.path.join(self._path, key),
                                        structure=subobj[key], parent=self))
            elif isinstance(subobj, MockPath):
                if subobj._parent is not None:
                    raise ValueError("attempting to add a child which already has a parent: {}".format(subobj))
                self.add_child(subobj)
            else:
                raise ValueError('items must be dict_like or MockPath: {}'.format(subobj))

    def _get_path(self):
        return self._path

    path = property(_get_path)

    def _get_parent(self):
        if self._parent is None:
            path = MockPath('subroot')
            path._children = [self]
            return path
        else:
            return self._parent

    def _set_parent(self, parent):
        if parent is None:
            self._parent = None
            parentpath = ''
        else:
            parentpath = parent.path
        self._parent = parent
        self._path = os.path.join(parentpath, self.name)
        for child in self.children:
            child.parent = self

    parent = property(_get_parent, _set_parent)

    def _get_name(self):
        return self._name

    def _set_name(self, name):
        self._name = name
        if self._parent is not None:
            self._path = os.path.join(self.parent.path, name)

    name = property(_get_name, _set_name)

    def _get_children(self):
        return self._children[:]

    children = property(_get_children)

    def _get_content(self):
        return self._content[:]

    file_content = property(_get_content)

    def __getitem__(self, name):
        """

        Parameters
        ----------
        name: str or list of strings

        Returns
        -------

        """
        next = []
        if isinstance(name, basestring):
            name = self._splitall(name)
        if isinstance(name, list):
            if len(name) == 1:
                name = name[0]
            else:
                next = name[1:]
                name = name[0]
        if isinstance(name, basestring):
            for child in self.iterdir():
                if child.name == name:
                    if next:
                        return child[next]
                    else:
                        return child
            raise KeyError("no name: {}".format(name))
        else:
            raise ValueError("name not a list or str: {}".format(name))

    def __contains__(self, name):
        """

        Parameters
        ----------
        name: str or list of strings

        Returns
        -------

        """
        try:
            self[name]
            return True
        except KeyError:
            return False

    def add_child(self, child):
        # TODO could allow same name if one is file and one is dir?
        if child.name in [c.name for c in self._children]:
            raise IOError("child with this name already exists: {}".format(child.name))
        child.parent = self
        self._children.append(child)

    # TODO need to implement relative naming and switchin to/from
    def absolute(self):
        return self

    # TODO should return a new mock path rather than a str, need to implement relative naming and switchin to/from
    def relative_to(self, other):
        return os.path.relpath(self._path, other._path)

    def rename(self, target):

        if hasattr(target, "name"):
            name = target.name
        elif isinstance(target, basestring):
            name = target
        else:
            raise ValueError("target must be a string or have a name attribute")

        if not self.exists():
            raise IOError("path doesn't exist: {}".format(self))
        self.name = name
        self._path = os.path.join(os.path.dirname(self._path), name)

    def _recurse_structure(self):
        structure = []
        for child in self.children:
            if not child.exists():
                continue
            if child.is_file():
                structure.append(child.copy_path_obj())
            else:
                structure.append({child.name: [c.copy_path_obj() for c in child.children]})
        return structure

    def copy_path_obj(self):
        """copy mock path (removing path and parent)"""
        if self.is_file():
            return MockPath(path=self.name, is_file=True,
                            exists=self.exists(), structure=[], content="\n".join(self._content), parent=None)
        else:
            structure = self._recurse_structure()
            return MockPath(path=self.name, is_file=False, exists=self.exists(), parent=None,
                            structure=structure)

    def is_file(self):
        return self._is_file

    def is_dir(self):
        return self._is_dir

    def exists(self):
        return self._exists

    def samefile(self, other):
        return self == other

    def _splitall(self, path):

        if isinstance(path, MockPath):
            path = os.path.relpath(str(path), str(self))
        if not isinstance(path, basestring):
            raise ValueError("path is not a string or MockPath: {}".format(path))

        allparts = []
        while 1:
            parts = os.path.split(path)
            if parts[0] == path:  # sentinel for absolute paths
                allparts.insert(0, parts[0])
                break
            elif parts[1] == path:  # sentinel for relative paths
                allparts.insert(0, parts[1])
                break
            else:
                path = parts[0]
                allparts.insert(0, parts[1])
        return allparts

    def _flatten(self, l):
        for el in l:
            if isinstance(el, collections.Iterable) and not isinstance(el, basestring):
                for sub in self._flatten(el):
                    yield sub
            else:
                yield el

    def joinpath(self, *paths):

        parts = list(self._flatten([self._splitall(path) for path in paths]))

        if not parts:
            raise IOError("must be at least one path")
        elif len(parts) == 1:

            if parts[0] == '' or parts[0] == '.':
                return self

            for child in self._children:
                if child.name == parts[0]:
                    return child

            # does not yet exist, must use touch or mkdir to convert to file or folder
            new = MockPath(path=os.path.join(self._path, parts[0]), exists=False, parent=self)
            self.add_child(new)
            return new

        else:
            for child in self._children:
                if child.name == parts[0]:
                    return child.joinpath(*parts[1:])
            new = MockPath(path=os.path.join(self._path, parts[0]), exists=False, parent=self)
            self.add_child(new)
            return new.joinpath(*parts[1:])

    # TODO add mode=0o777, exist_ok=False
    def mkdir(self, parents=False):
        """

        Parameters
        ----------
        mode
        parents: bool
            If parents is true, any missing parents of this path are created as needed
            If parents is false, a missing parent raises FileNotFoundError.

        Returns
        -------

        """
        if self.parent is not None:
            if not self.parent.exists():
                if not parents:
                    raise FileNotFoundError("the parent must exist for {}".format(self))
                else:
                    self.parent.mkdir(parents=parents)  # mode=0o777,

        if not self._exists:
            self._is_file = False
            self._is_dir = True
            self._exists = True

    # TODO store stat attributes and apply them to mktemp
    def stat(self):
        """ Retrieve information about a file

        Parameters
        ----------
        path: str

        Returns
        -------
        attr: object
            see os.stat, includes st_mode, st_size, st_uid, st_gid, st_atime, and st_mtime attributes

        """
        class MockStat(object):
            # at present just returning a typical file result
            def __init__(self):
                self.st_mode = 33188
                self.st_ino = 74204932
                self.st_dev = 16777220
                self.st_nlink = 1
                self.st_uid = 634541
                self.st_gid = 1335817362
                self.st_size = 10410
                self.st_atime = 1504518028
                self.st_mtime = 1504518028
                self.st_ctime = 1504518028

            def __repr__(self):
                return "MockStatResult(st_mode=33188, st_ino=74204932, st_dev=16777220, st_nlink=1, st_uid=634541, " \
                    "st_gid=1335817362, st_size=10410, st_atime=1504518028, st_mtime=1504518028, st_ctime=1504518028)"

        return MockStat()

    def chmod(self, mode):
        """ Change the mode (permissions) of a file

        Parameters
        ----------
        path: str
        mode: int
            new permissions (see os.chmod)

        Examples
        --------
        To make a file executable
        cur_mode = folder.stat("exec.sh").st_mode
        folder.chmod("exec.sh", cur_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH )

        """
        pass

    def touch(self):

        if self.parent is not None:
            if not self.parent.exists():
                raise IOError("the parent must exist")

        if not self._exists:
            self._is_file = True
            self._is_dir = False
            self._exists = True

    def unlink(self):
        if not self._is_file:
            raise IOError("path is not a file")
        self._exists = False

    def rmdir(self):
        if not self._is_dir:
            raise IOError("path is not a directory")
        if list(self.iterdir()):
            raise IOError("path is not empty: {}".format(list(self.iterdir())))
        self._exists = False

    def iterdir(self):
        for subobj in sorted(self.children):
            if subobj.exists():
                yield subobj

    # TODO if glob("**/stop") should stop recursing once found
    def glob(self, regex, recurse=False, toplevel=True):
        """

        Parameters
        ----------
        regex: str
            the path regex, with * to match 0 or more (non-recursive) paths and ** to match zero or more (recursive) directories
        recurse: bool

        Yields
        -------
        path: MockPath

        """
        if not regex and toplevel:
            raise ValueError("Unacceptable pattern: ''")
        parts = self._splitall(regex)
        if parts[0] == "**":
            recurse = True
            parts = [""] if len(parts) == 1 else parts[1:]
        if len(parts) == 1:
            for subobj in sorted(self.iterdir()):
                if subobj.is_dir() and recurse:
                    yield subobj
                    for path in subobj.glob(parts[0], recurse=recurse, toplevel=False):
                        yield path
                elif fnmatch(subobj.name, parts[0]):
                    yield subobj
        else:
            for subobj in sorted(self.iterdir()):
                if subobj.is_dir() and recurse:
                    yield subobj
                    for path in subobj.glob(os.path.join(*parts[1:]), recurse=recurse, toplevel=False):
                        yield path
                elif fnmatch(subobj.name, parts[0]):
                    for path in subobj.glob(os.path.join(*parts[1:]), recurse=recurse, toplevel=False):
                        yield path

    @contextlib.contextmanager
    def maketemp(self, getoutput=False, dir=None):
        """make a named temporary file or folder containing the path contents

        Parameters
        ----------
        getoutput: bool
            if True, (on exit) new paths will be read/added to the path
        dir: None or str
            directory to place temp in (see tempfile.mkstemp)

        Yields
        ------
        temppath: path.Path
            path to temporary

        """
        if self.is_file():
            filetemp = tempfile.NamedTemporaryFile(mode='w+', delete=False, dir=dir)
            try:
                filetemp.write('\n'.join(self._content))
                filetemp.close()
                dirpath = os.path.join(os.path.dirname(filetemp.name), self.name)
                os.rename(filetemp.name, dirpath)
                yield pathlib.Path(dirpath)
            finally:
                try:
                    if getoutput:
                        raise NotImplementedError
                finally:
                    os.remove(dirpath)
        else:
            temppath = pathlib.Path(tempfile.mkdtemp(dir=dir))
            dirpath = os.path.join(os.path.dirname(str(temppath)), self.name)
            os.rename(str(temppath), dirpath)
            temppath = pathlib.Path(dirpath)
            try:
                self.copy_to(temppath.parent)
                yield pathlib.Path(dirpath)
            finally:
                temppath = pathlib.Path(dirpath)
                try:
                    if getoutput:
                        self._iter_temp(self, temppath, overwrite=False)
                finally:
                    for subpath in temppath.glob("**/*"):
                        if subpath.is_file():
                            subpath.unlink()
                    for subpath in reversed(list(temppath.glob("**"))):
                        subpath.rmdir()
                    if temppath.exists():
                        temppath.rmdir()

    def _iter_temp(self, mock, temp, overwrite=False):
        if not mock.name == temp.name:
            raise ValueError("mock name and temp name different: {0}, {1}".format(mock.name, temp.name))
        child_names = {c.name: c for c in mock.children}
        for subpath in temp.iterdir():
            if subpath.is_file():
                if subpath.name in child_names and not overwrite:
                    pass
                else:
                    with subpath.open() as f:
                        newpath = MockPath(subpath.name, is_file=True, content=f.read(), exists=True)
                    mock.add_child(newpath)
            else:
                if subpath.name not in child_names:
                    newpath = MockPath(subpath.name, exists=True)
                    mock.add_child(newpath)
                else:
                    newpath = child_names[subpath.name]
                self._iter_temp(newpath, subpath, overwrite=overwrite)

    def copy_to(self, target):
        """ copy from a mock path to a target

        Parameters
        ----------
        target: str, pathlib.Path or MockPath

        Returns
        -------

        """
        if not self.exists():
            raise IOError("this path does not exist")

        if isinstance(target, basestring):
            target = pathlib.Path(target)
        if isinstance(target, pathlib.Path):
            if not target.is_dir():
                raise IOError("target is not a directory")
            ignore = len(self.path) - len(self.name)
            for path in self.glob("**/*"):
                newpath = target.joinpath(str(path)[ignore:])
                if path.is_dir():
                    if not newpath.exists():
                        newpath.mkdir()
                else:
                    if newpath.exists():
                        raise IOError("file already exists: {}".format(newpath))
                    else:
                        newpath.touch()
                    with newpath.open('w') as f:
                        if sys.version_info.major > 2:
                            f.write("\n".join(path._content))
                        else:
                            f.write(unicode('\n'.join(path._content)))

        elif isinstance(target, MockPath):
            if not target.is_dir():
                raise IOError("target is not a directory")
            newpath = self.copy_path_obj()
            return target.add_child(newpath)
        else:
            raise ValueError("target is not str, pathlib.Path or MockPath: {}".format(target))

    def copy_from(self, source):
        """ copy from a source to a mock directory

        Parameters
        ----------
        source: str, file_obj, pathlib.Path or MockPath

        Returns
        -------

        """
        if not self.is_dir() or not self.exists():
            raise IOError("this path is not an existing directory")

        if isinstance(source, basestring):
            source = pathlib.Path(source)
            if not source.exists():
                raise IOError("source does not exist: {}".format(source))
        if isinstance(source, MockPath):
            if not source.exists():
                raise IOError("source does not exist: {}".format(source))
            else:
                newpath = source.copy_path_obj()
                self.add_child(newpath)
        elif isinstance(source, pathlib.Path):
            if not source.exists():
                raise IOError("source does not exist: {}".format(source))
            elif source.is_file():
                with source.open() as f:
                    content = f.read()
                newfile = MockPath(source.name, is_file=True, parent=self, content=content)
                self.add_child(newfile)
            else:
                raise NotImplementedError
        else:
            raise ValueError("source is not str, pathlib.Path or MockPath: {}".format(source))

    @contextlib.contextmanager
    def open(self, mode='r', encoding=None):
        """ context manager for opening a file

        Parameters
        ----------
        mode: str
        encoding: None or str

        Returns
        -------

        """
        if self.is_dir():
            raise IOError('[Errno 21] Is a directory: {}'.format(self.path))

        if 'r' in mode:
            obj = _OpenRead(self._content, encoding, "b" in mode)
            yield obj
        elif 'w' in mode:
            obj = _OpenWrite("b" in mode)
            yield obj
            self._content = obj._str.splitlines()
        else:
            raise ValueError('readwrite should contain r or w')

    def __gt__(self, other):
        if not hasattr(other, 'name'):
            return NotImplemented
        return self.path > other.path

    def __eq__(self, other):
        if not hasattr(other, 'name'):
            return NotImplemented
        return self.path == other.path

    def _recurse_print(self, obj, text='', indent=0, indentlvl=2, file_content=False):
        indent += indentlvl
        for subobj in sorted(obj):
            if not subobj.exists():
                continue
            if subobj.is_dir():
                text += ' ' * indent + '{0}("{1}")\n'.format(self._folderstr, subobj.name)
                text += self._recurse_print(subobj.iterdir(),
                                            indent=indent, file_content=file_content)
            else:
                if file_content:
                    sep = '\n' + ' ' * (indent + 1)
                    text += ' ' * indent + sep.join(
                        ['{0}("{1}") Contents:'.format(self._filestr, subobj.name)] + subobj._content) + '\n'
                else:
                    text += ' ' * indent + '{0}("{1}")\n'.format(self._filestr, subobj.name)

        return text

    def to_string(self, indentlvl=2, file_content=False, color=False):
        """convert to string """
        if color:
            self._folderstr = colortxt('Folder', 'green')
            self._filestr = colortxt('File', 'blue')
        else:
            self._folderstr = 'Folder'
            self._filestr = 'File'

        if self.is_file():
            return '\n'.join(['{0}("{1}") Contents:'.format(self._filestr, self.name)] + self._content)
        elif self.is_dir():
            text = '{0}("{1}")\n'.format(self._folderstr, self.name)
            text += self._recurse_print(self.iterdir(), indentlvl=indentlvl,
                                        file_content=file_content)

            text = text[0:-1] if text.endswith('\n') else text
            return text
        else:
            return 'MockPath({})'.format(self.name)

    def __str__(self):
        return self._path  # self.__repr__()

    def __repr__(self):
        if not self.exists():
            return 'MockVirtualPath("{}")'.format(self._path)
        elif self.is_dir():
            return 'MockFolder("{}")'.format(self._path)
        elif self.is_file():
            return 'MockFile("{}")'.format(self._path)
        else:
            return 'MockPath("{}")'.format(self._path)

_ATTRIBUTES = dict(
    list(zip([
        'bold',
        'dark',
        '',
        'underline',
        'blink',
        '',
        'reverse',
        'concealed'
    ],
        list(range(1, 9))
    ))
)
del _ATTRIBUTES['']

_HIGHLIGHTS = dict(
    list(zip([
        'on_grey',
        'on_red',
        'on_green',
        'on_yellow',
        'on_blue',
        'on_magenta',
        'on_cyan',
        'on_white'
    ],
        list(range(40, 48))
    ))
)

_COLORS = dict(
    list(zip([
        'grey',
        'red',
        'green',
        'yellow',
        'blue',
        'magenta',
        'cyan',
        'white',
    ],
        list(range(30, 38))
    ))
)


def colortxt(text, color=None, on_color=None, attrs=None):
    """Colorize text.

    Available text colors:
        red, green, yellow, blue, magenta, cyan, white.

    Available text highlights:
        on_red, on_green, on_yellow, on_blue, on_magenta, on_cyan, on_white.

    Available attributes:
        bold, dark, underline, blink, reverse, concealed.

    Examples
    --------
    >>> txt = colortxt('Hello, World!', 'red', 'on_grey', ['bold'])
    >>> print(txt)
    \x1b[1m\x1b[40m\x1b[31mHello, World!\x1b[0m

    """
    _RESET = '\033[0m'
    __ISON = True
    if __ISON and os.getenv('ANSI_COLORS_DISABLED') is None:
        fmt_str = '\033[%dm%s'
        if color is not None:
            text = fmt_str % (_COLORS[color], text)

        if on_color is not None:
            text = fmt_str % (_HIGHLIGHTS[on_color], text)

        if attrs is not None:
            for attr in attrs:
                text = fmt_str % (_ATTRIBUTES[attr], text)

        text += _RESET
    return text



"""
https://stackoverflow.com/questions/1960516/python-json-serialize-a-decimal-object
"""
from decimal import Decimal


class Encode_Decimal(object):
    """

    Examples
    --------
    >>> from decimal import Decimal
    >>> Encode_Decimal().to_str(Decimal('1.2345'))
    '1.2345'
    >>> Encode_Decimal().to_json(Decimal('1.2345'))
    {'_python_Decimal_': '1.2345'}
    >>> Encode_Decimal().from_json({'_python_Decimal_': '1.2345'})
    Decimal('1.2345')

    """

    plugin_name = 'decimal.Decimal'
    plugin_descript = 'encode/decode Decimal type'
    objclass = Decimal
    dict_signature = ['_python_Decimal_']

    def to_str(self, obj):
        return obj.to_eng_string()

    def to_json(self, obj):
        return {'_python_Decimal_': obj.to_eng_string()}

    def from_json(self, obj):
        return Decimal(obj['_python_Decimal_'])



#!/usr/bin/env python
# -- coding: utf-8 --

## TODO see below
# output json with arrays all on one line
#      doesn't seem an easy way
#      could use this: https://stackoverflow.com/a/13252112, but wrap lists with NoIndent first
#      would need to go through first and wrap values that are list instances with NoIndent first


# internal packages
import json
import os
from decimal import Decimal

# python 3 to 2 compatibility
try:
    import pathlib
except ImportError:
    import pathlib2 as pathlib
try:
    basestring
except NameError:
    basestring = str
try:
    from urllib2 import urlopen
except ImportError:
    from urllib.request import urlopen

# external packages
import warnings

warnings.simplefilter('once', ImportWarning)
try:
    import ijson
except ImportError:
    pass

# local imports
from jsonextended.edict import indexes, pprint, convert_type
from jsonextended.plugins import decode


def _get_keys(file_obj, key_path=None):
    key_path = [] if key_path is None else key_path
    data = json.load(file_obj, object_hook=decode)
    data = indexes(data, key_path)
    if hasattr(data, 'keys'):
        return sorted([str(k) if isinstance(k, basestring) else k for k in data.keys()])
    else:
        return []


def _get_keys_ijson(file_obj, key_path=None):
    key_path = [] if key_path is None else key_path
    try:
        path_str = '.'.join(key_path)
        keys = []
        for prefix, etype, value in ijson.parse(file_obj):
            if etype == 'map_key':
                if prefix == path_str:
                    keys.append(value)
        return sorted([str(k) if isinstance(k, basestring) else k for k in keys])
    except NameError:
        warnings.warn('ijson package not found in environment, \
please install for on-disk key indexing', ImportWarning)
        return _get_keys(file_obj, key_path)


def _get_keys_folder(jdir, key_path=None, in_memory=True, ignore_prefix=('.', '_')):
    """ get json keys from directory structure

    e.g.

    jdir
        sub_dir1
            data1.json
            data2.json
        sub_dir2
            data.json

    _get_keys_folder(jdir)
    => ['sub_dir1', 'sub_dir2']
    _get_keys_folder(jdir,['sub_dir1'])
    => ['data1', 'data2']

    NB: json files are identified with .json extension
        files/directories beginning with '.' are ignored

    """
    if not hasattr(jdir, 'iterdir'):
        raise ValueError('jdir is not a path object; {}'.format(jdir))

    key_path = [] if key_path is None else key_path

    keys = []

    key_found = False if key_path else True
    search_key = key_path[0] if len(key_path) > 0 else None

    for jsub in jdir.iterdir():
        if jsub.is_file() and jsub.name[-5:] == '.json':

            name, ext = os.path.splitext(jsub.name)
            if name == search_key or not key_path:
                key_found = True
                if key_path:
                    return jkeys(jsub, key_path[1:], in_memory, ignore_prefix)
                else:
                    keys.append(name)

        elif jsub.is_dir() and not jsub.name.startswith(ignore_prefix) and (jsub.name == search_key or not key_path):

            key_found = True
            if jsub.name in keys:
                raise IOError(
                    'directory has a sub-dir and file with same name: {1} and {1}.json in {0}'.format(jdir, jsub.name))
            if key_path:
                return jkeys(jsub, key_path[1:], in_memory, ignore_prefix)
            else:
                keys.append(jsub.name)

    if not key_found:
        raise KeyError('key not found: {0}'.format(search_key))

    return sorted(keys)


def jkeys(jfile, key_path=None, in_memory=True, ignore_prefix=('.', '_')):
    """ get keys for initial json level, or at level after following key_path

    Parameters
    ----------
    jfile : str, file_like or path_like
        if str, must be existing file or folder,
        if file_like, must have 'read' method
        if path_like, must have 'iterdir' method (see pathlib.Path)
    key_path : list of str
        a list of keys to index into the json before returning keys
    in_memory : bool
        if true reads json into memory before finding keys (this is faster but uses more memory)
    ignore_prefix : list of str
        ignore folders beginning with these prefixes

    Examples
    --------

    >>> from jsonextended.utils import MockPath
    >>> file_obj = MockPath('test.json',is_file=True,
    ... content='''
    ... {
    ...  "a": 1,
    ...  "b": [1.1,2.1],
    ...  "c": {"d":"e","f":"g"}
    ... }
    ... ''')
    ...
    >>> jkeys(file_obj)
    ['a', 'b', 'c']

    >>> jkeys(file_obj,["c"])
    ['d', 'f']

    >>> from jsonextended.utils import get_test_path
    >>> path = get_test_path()
    >>> jkeys(path)
    ['dir1', 'dir2', 'dir3']

    >>> path = get_test_path()
    >>> jkeys(path, ['dir1','file1'], in_memory=True)
    ['initial', 'meta', 'optimised', 'units']

    """
    key_path = [] if key_path is None else key_path

    def eval_file(file_obj):
        if not in_memory:
            return _get_keys_ijson(file_obj, key_path)
        else:
            return _get_keys(file_obj, key_path)

    if isinstance(jfile, basestring):
        if not os.path.exists(jfile):
            raise IOError('jfile does not exist: {}'.format(jfile))
        if os.path.isdir(jfile):
            jpath = pathlib.Path(jfile)
            return _get_keys_folder(jpath, key_path, in_memory, ignore_prefix)
        else:
            with open(jfile, 'r') as file_obj:
                return eval_file(file_obj)
    elif hasattr(jfile, 'read'):
        return eval_file(jfile)
    elif hasattr(jfile, 'iterdir'):
        if jfile.is_file():
            with jfile.open('r') as file_obj:
                return eval_file(file_obj)
        else:
            return _get_keys_folder(jfile, key_path, in_memory, ignore_prefix)
    else:
        raise ValueError('jfile should be a str, file_like or path_like object: {}'.format(jfile))


def _file_with_keys(file_obj, key_path=None, parse_decimal=False):
    """read json with keys

    Parameters
    ----------
    file_obj : object
        object with read method
    key_path : list of str
        key to index befor parsing
    parse_decimal : bool
        whether to parse numbers as Decimal instances (retains exact precision)

    Notes
    -----
    ijson outputs decimals as Decimal class (for arbitrary precision)

    """
    key_path = [] if key_path is None else key_path

    try:
        objs = ijson.items(file_obj, '.'.join(key_path))
    except NameError:
        warnings.warn('ijson package not found in environment, \
        please install for on-disk key indexing', ImportWarning)
        data = json.load(file_obj, parse_float=Decimal if parse_decimal else float, object_hook=decode)
        return indexes(data, key_path)
    try:
        data = next(objs)  # .next()
    except StopIteration:
        raise KeyError('key path not available in json: {}'.format(key_path))

    # by default ijson parses Decimal values
    if not parse_decimal:
        convert_type(data, Decimal, float, in_place=True)

    datastr = json.dumps(data)
    data = json.loads(datastr, object_hook=decode)

    return data


# TODO this is a hack to get _folder_to_json to work if last key_path is at a leaf node, should improve
class _Terminus(object):
    def __hash__(self):
        return 1

    def __eq__(self, other):
        return True


def _folder_to_json(jdir, key_path=None, in_memory=True,
                    ignore_prefix=('.', '_'), dic={}, parse_decimal=False):
    """ read in folder structure as json

    e.g.

    jdir
        sub_dir1
            data.json
        sub_dir2
            data.json

    _folder_to_json(jdir)
    => {'sub_dir1':{'data':{...}},
        'sub_dir2':{'data':{...}}}

    NB: json files are identified with .json extension

    """
    key_path = [] if key_path is None else key_path

    if not hasattr(jdir, 'iterdir'):
        raise ValueError('jdir is not a path object; {}'.format(jdir))

    key_found = False if key_path else True
    search_key = key_path[0] if len(key_path) > 0 else None

    for jsub in jdir.iterdir():
        if jsub.is_file() and jsub.name.endswith('.json'):

            name, ext = os.path.splitext(jsub.name)
            if name == search_key or not key_path:
                key_found = True
                if key_path:
                    data = to_dict(jsub, key_path[1:], in_memory, ignore_prefix, parse_decimal)
                    if isinstance(data, dict):
                        dic.update(data)
                    else:
                        dic.update({_Terminus(): data})
                else:
                    dic[name] = to_dict(jsub, key_path[1:], in_memory, ignore_prefix, parse_decimal)

        elif jsub.is_dir() and not jsub.name.startswith(ignore_prefix) and (jsub.name == search_key or not key_path):

            key_found = True
            if jsub.name in dic.keys():
                raise IOError(
                    'directory has a sub-dir and file with same name: {1} and {1}.json in {0}'.format(jdir, jsub.name))
            if key_path:
                sub_d = dic
            else:
                dic[jsub.name] = {}
                sub_d = dic[jsub.name]
            _folder_to_json(jsub, key_path[1:], in_memory, ignore_prefix, sub_d, parse_decimal)

    if not key_found:
        raise KeyError('key not found: {0}'.format(search_key))


def to_dict(jfile, key_path=None, in_memory=True,
            ignore_prefix=('.', '_'), parse_decimal=False):
    """ input json to dict

    Parameters
    ----------
    jfile : str, file_like or path_like
        if str, must be existing file or folder,
        if file_like, must have 'read' method
        if path_like, must have 'iterdir' method (see pathlib.Path)
    key_path : list of str
        a list of keys to index into the json before parsing it
    in_memory : bool
        if true reads full json into memory before filtering keys (this is faster but uses more memory)
    ignore_prefix : list of str
        ignore folders beginning with these prefixes
    parse_decimal : bool
        whether to parse numbers as Decimal instances (retains exact precision)

    Examples
    --------

    >>> from pprint import pformat

    >>> from jsonextended.utils import MockPath
    >>> file_obj = MockPath('test.json',is_file=True,
    ... content='''
    ... {
    ...  "a": 1,
    ...  "b": [1.1,2.1],
    ...  "c": {"d":"e"}
    ... }
    ... ''')
    ...

    >>> dstr = pformat(to_dict(file_obj))
    >>> print(dstr.replace("u'","'"))
    {'a': 1, 'b': [1.1, 2.1], 'c': {'d': 'e'}}

    >>> dstr = pformat(to_dict(file_obj,parse_decimal=True))
    >>> print(dstr.replace("u'","'"))
    {'a': 1, 'b': [Decimal('1.1'), Decimal('2.1')], 'c': {'d': 'e'}}

    >>> str(to_dict(file_obj,["c","d"]))
    'e'

    >>> from jsonextended.utils import get_test_path
    >>> path = get_test_path()
    >>> jdict1 = to_dict(path)
    >>> pprint(jdict1,depth=2)
    dir1:
      dir1_1: {...}
      file1: {...}
      file2: {...}
    dir2:
      file1: {...}
    dir3:

    >>> jdict2 = to_dict(path,['dir1','file1','initial'],in_memory=False)
    >>> pprint(jdict2,depth=1)
    crystallographic: {...}
    primitive: {...}

    """
    key_path = [] if key_path is None else key_path

    if isinstance(jfile, basestring):
        if not os.path.exists(jfile):
            raise IOError('jfile does not exist: {}'.format(jfile))
        if os.path.isdir(jfile):
            data = {}
            jpath = pathlib.Path(jfile)
            _folder_to_json(jpath, key_path[:], in_memory, ignore_prefix, data, parse_decimal)
            if isinstance(list(data.keys())[0], _Terminus):
                data = data.values()[0]
        else:
            with open(jfile, 'r') as file_obj:
                if key_path and not in_memory:
                    data = _file_with_keys(file_obj, key_path, parse_decimal)
                elif key_path:
                    data = json.load(file_obj, parse_float=Decimal if parse_decimal else float, object_hook=decode)
                    data = indexes(data, key_path)
                else:
                    data = json.load(file_obj, parse_float=Decimal if parse_decimal else float, object_hook=decode)
    elif hasattr(jfile, 'read'):
        if key_path and not in_memory:
            data = _file_with_keys(jfile, key_path, parse_decimal)
        elif key_path:
            data = json.load(jfile, parse_float=Decimal if parse_decimal else float, object_hook=decode)
            data = indexes(data, key_path)
        else:
            data = json.load(jfile, parse_float=Decimal if parse_decimal else float, object_hook=decode)
    elif hasattr(jfile, 'iterdir'):
        if jfile.is_file():
            with jfile.open() as file_obj:
                if key_path and not in_memory:
                    data = _file_with_keys(file_obj, key_path, parse_decimal)
                elif key_path:
                    data = json.load(file_obj, parse_float=Decimal if parse_decimal else float, object_hook=decode)
                    data = indexes(data, key_path)
                else:
                    data = json.load(file_obj, parse_float=Decimal if parse_decimal else float, object_hook=decode)
        else:
            data = {}
            _folder_to_json(jfile, key_path[:], in_memory, ignore_prefix, data, parse_decimal)
            if isinstance(list(data.keys())[0], _Terminus):
                data = data.values()[0]
    else:
        raise ValueError('jfile should be a str, file_like or path_like object: {}'.format(jfile))

    return data



#!/usr/bin/env python

import json


class JSON_Parser(object):
    """

    Examples
    --------

    >>> from jsonextended.utils import MockPath
    >>> fileobj = MockPath(is_file=True,
    ... content='{"key1":[1,2,3]}'
    ... )
    >>> with fileobj.open() as f:
    ...     data = JSON_Parser().read_file(f)
    >>> list(data.values())
    [[1, 2, 3]]

    """

    plugin_name = 'json.basic'
    plugin_descript = 'read *.json files using json.load'
    file_regex = '*.json'

    def read_file(self, file_obj, **kwargs):
        return json.load(file_obj, object_hook=kwargs.get('object_hook', None),
                         parse_float=kwargs.get('parse_float', None))



"""
https://stackoverflow.com/questions/1960516/python-json-serialize-a-decimal-object
"""
from fractions import Fraction


class Encode_Fraction(object):
    """

    Examples
    --------
    >>> from decimal import Decimal
    >>> Encode_Fraction().to_str(Fraction(1, 3))
    '1/3'
    >>> Encode_Fraction().to_json(Fraction('1/3'))
    {'_python_Fraction_': '1/3'}
    >>> Encode_Fraction().from_json({'_python_Fraction_': '1/3'})
    Fraction(1, 3)

    """

    plugin_name = 'fractions.Fraction'
    plugin_descript = 'encode/decode Fraction type'
    objclass = Fraction
    dict_signature = ['_python_Fraction_']

    def to_str(self, obj):
        return str(obj)

    def to_json(self, obj):
        return {'_python_Fraction_': str(obj)}

    def from_json(self, obj):
        return Fraction(obj['_python_Fraction_'])



class Encode_Set(object):
    """

    Examples
    --------

    >>> Encode_Set().to_str(set([1,2,3]))
    '{1, 2, 3}'

    >>> Encode_Set().to_json(set([1,2,3]))
    {'_python_set_': [1, 2, 3]}

    >>> list(Encode_Set().from_json({'_python_set_': [1, 2, 3]}))
    [1, 2, 3]

    """

    plugin_name = 'python.set'
    plugin_descript = 'decode/encode python set'
    objclass = set
    dict_signature = ['_python_set_']

    def to_str(self, obj):
        return '{' + str(list(obj))[1:-1] + '}'

    def to_json(self, obj):
        return {'_python_set_': list(obj)}

    def from_json(self, obj):
        return set(obj['_python_set_'])



"""
https://stackoverflow.com/questions/27909658/json-encoder-and-decoder-for-complex-numpy-arrays
"""

import numpy as np

try:
    from functools import reduce
except ImportError:
    pass
import operator


class Encode_NDArray(object):
    """

    Examples
    --------
    >>> from pprint import pprint
    >>> import numpy as np

    >>> Encode_NDArray().to_str(np.asarray([1,2,3]))
    '[1 2 3]'

    >>> pprint(Encode_NDArray().to_json(np.asarray([1,2,3])))
    {'_numpy_ndarray_': {'dtype': 'int64', 'value': [1, 2, 3]}}

    >>> Encode_NDArray().from_json({'_numpy_ndarray_': {'dtype': 'int64', 'value': [1, 2, 3]}})
    array([1, 2, 3])

    """

    plugin_name = 'numpy.ndarray'
    plugin_descript = 'encode/decode numpy.ndarray'
    objclass = np.ndarray
    dict_signature = ['_numpy_ndarray_']

    def to_str(self, obj):
        elements = reduce(operator.mul, obj.shape, 1)
        if elements > 10:
            return 'np.array({0}, min={1:.2E}, max={2:.2E})'.format(
                obj.shape, np.nanmin(obj), np.nanmax(obj))
        else:
            return ' '.join(str(obj).split())

    def to_json(self, obj):
        return {'_numpy_ndarray_': {'value': obj.tolist(), 'dtype': str(obj.dtype)}}

    def from_json(self, obj):
        return np.array(obj['_numpy_ndarray_']['value'],
                        dtype=obj['_numpy_ndarray_']['dtype'])



#!/usr/bin/env python

import h5py


class HDF5_Parser(object):
    """

    Examples
    --------

    >>> import h5py
    >>> indata = h5py.File('test.hdf5')
    >>> dataset = indata.create_dataset("mydataset", (10,), dtype='i')
    >>> indata.close()

    >>> with open('test.hdf5') as f:
    ...     data = HDF5_Parser().read_file(f)
    >>> data['mydataset'][:]
    array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0], dtype=int32)

    >>> import os
    >>> os.remove('test.hdf5')

    """

    plugin_name = 'hdf5.read'
    plugin_descript = 'read *.hdf5 (in read mode) files using h5py'
    file_regex = '*.hdf5'

    def read_file(self, file_obj, **kwargs):
        return h5py.File(file_obj.name, mode='r')



#!/usr/bin/env python
# -- coding: utf-8 --
""" a module to extend the python json package functionality;

-  decoding/encoding between the on-disk JSON structure
   and in-memory nested dictionary structure, including:

   -  treating path structures, with nested directories and multiple .json files, as a single json.

   -  on-disk indexing of the json structure (using the ijson package)

   -  extended data type serialisation (numpy.ndarray, Decimals,
      pint.Quantities,...)

-  viewing and manipulating the nested dictionaries:

   -  enhanced pretty printer

   -  Javascript rendered, expandable tree in the Jupyter Notebook

   -  filter, merge, flatten, unflatten functions

-  Units schema concept to apply and convert physical units (using the
   pint package)

-  Parser abstract class for dealing with converting other file formats
   to JSON

Notes
-----

On-disk indexing of the json structure, before reading into memory,
to reduce memory overhead when dealing with large json structures/files (using the ijson package).
e.g.
    path = get_test_path()
    %memit jdict1 = to_dict(path,['dir1','file2','meta'],in_memory=True)
    maximum of 3: 12.242188 MB per loop

    %memit jdict1 = to_dict(path,['dir1','file2','meta'],in_memory=False)
    maximum of 3: 6.996094 MB per loop


Examples
--------

>>> from jsonextended import ejson, edict, utils

>>> path = utils.get_test_path()
>>> path.is_dir()
True

>>> ejson.jkeys(path)
['dir1', 'dir2', 'dir3']

>>> jdict1 = ejson.to_dict(path)
>>> edict.pprint(jdict1,depth=2)
dir1:
  dir1_1: {...}
  file1: {...}
  file2: {...}
dir2:
  file1: {...}
dir3:


>>> jdict2 = ejson.to_dict(path,['dir1','file1'])
>>> edict.pprint(jdict2,depth=1)
initial: {...}
meta: {...}
optimised: {...}
units: {...}

>>> filtered = edict.filter_keys(jdict2,['vol*'],use_wildcards=True)
>>> edict.pprint(filtered)
initial:
  crystallographic:
    volume: 924.62752781
  primitive:
    volume: 462.313764
optimised:
  crystallographic:
    volume: 1063.98960509
  primitive:
    volume: 531.994803

>>> edict.pprint(edict.flatten(filtered))
(initial, crystallographic, volume):   924.62752781
(initial, primitive, volume):          462.313764
(optimised, crystallographic, volume): 1063.98960509
(optimised, primitive, volume):        531.994803

"""

__version__ = '0.7.8'

from jsonextended import ejson, units, utils, edict, plugins


def _run_nose_tests(doctests=True, verbose=True):
    """
    mimics nosetests --with-doctest -v --exe jsonextended
    also use:
    pylint --output-format html jsonextended > jsonextended_pylint.html
    """
    import os, sys, jsonextended, nose
    nose_argv = sys.argv
    nose_argv += ['--detailed-errors', '--exe']
    if verbose:
        nose_argv.append('-v')
    if doctests:
        nose_argv.append('--with-doctest')
    nose_argv.append('jsonextended')
    initial_dir = os.getcwd()
    my_package_file = os.path.abspath(jsonextended.__file__)
    print(my_package_file)
    my_package_dir = os.path.dirname(os.path.dirname(my_package_file))
    print(my_package_dir)
    os.chdir(my_package_dir)
    try:
        nose.run(argv=nose_argv)
    finally:
        os.chdir(initial_dir)



#!/usr/bin/env python

class KeyPair_Parser(object):
    """
    Examples
    --------

    >>> from pprint import pprint
    >>> from jsonextended.utils import MockPath
    >>> fileobj = MockPath(is_file=True,
    ... content='''# comment line
    ... key1 val1
    ... key2 val2
    ... key3 val3'''
    ... )
    >>> with fileobj.open() as f:
    ...     data = KeyPair_Parser().read_file(f)
    >>> pprint(data)
    {'key1': 'val1', 'key2': 'val2', 'key3': 'val3'}

    """

    plugin_name = 'keypair'
    plugin_descript = "read *.keypair, where each line should be; '<key> <pair>'"
    file_regex = '*.keypair'

    def read_file(self, file_obj, **kwargs):

        comments = kwargs.get('comments', '#')
        delim = kwargs.get('keypair_delim', None)
        keypair = {}
        for line in file_obj:
            if line.strip().startswith(comments):
                continue
            key, pair = line.strip().split(delim)
            keypair[key] = pair
        return keypair



import inspect
import re
import os

# python 2/3 compatibility
try:
    basestring
except NameError:
    basestring = str
# python 3 to 2 compatibility
try:
    import pathlib
except ImportError:
    import pathlib2 as pathlib

# back compatibility
from jsonextended.mockpath import MockPath, colortxt

def class_to_str(obj):
    """ get class string from object

    Examples
    --------

    >>> class_to_str(list).split('.')[1]
    'list'

    """
    mod_str = obj.__module__
    name_str = obj.__name__
    if mod_str == '__main__':
        return name_str
    else:
        return '.'.join([mod_str, name_str])


def get_module_path(module):
    """return a directory path to a module"""
    return pathlib.Path(os.path.dirname(os.path.abspath(inspect.getfile(module))))


from jsonextended import _example_data_folder


def get_test_path():
    """ returns test path object

    Examples
    --------
    >>> path = get_test_path()
    >>> path.name
    '_example_data_folder'

    """
    return get_module_path(_example_data_folder)


def get_data_path(data, module, check_exists=True):
    """return a directory path to data within a module

    data : str or list of str
        file name or list of sub-directories and file name (e.g. ['lammps','data.txt'])
    """
    basepath = os.path.dirname(os.path.abspath(inspect.getfile(module)))

    if isinstance(data, basestring): data = [data]

    dirpath = os.path.join(basepath, *data)

    if check_exists:
        assert os.path.exists(dirpath), '{0} does not exist'.format(dirpath)

    return pathlib.Path(dirpath)


def _atoi(text):
    return int(text) if text.isdigit() else text


def _natural_keys(text):
    """human order sorting

    alist.sort(key=_natural_keys)
    """
    return [_atoi(c) for c in re.split('(\d+)', str(text))]


def natural_sort(iterable):
    """human order sorting of number strings

    Examples
    --------

    >>> sorted(['011','1', '21'])
    ['011', '1', '21']

    >>> natural_sort(['011','1', '21'])
    ['1', '011', '21']

    """
    return sorted(iterable, key=_natural_keys)


def memory_usage():
    """return memory usage of python process in MB

    from http://fa.bianp.net/blog/2013/different-ways-to-get-memory-consumption-or-lessons-learned-from-memory_profiler/
    psutil is quicker

    >>> isinstance(memory_usage(),float)
    True

    """
    try:
        import psutil, os
    except ImportError:
        return _memory_usage_ps()

    process = psutil.Process(os.getpid())
    mem = process.memory_info()[0] / float(2 ** 20)
    return mem


def _memory_usage_ps():
    """return memory usage of python process in MB

    >>> isinstance(_memory_usage_ps(),float)
    True

    """
    import subprocess, os
    out = subprocess.Popen(['ps', 'v', '-p', str(os.getpid())],
                           stdout=subprocess.PIPE).communicate()[0].split(b'\n')
    vsz_index = out[0].split().index(b'RSS')
    mem = float(out[1].split()[vsz_index]) / 1024
    return mem


def load_memit():
    """load memory usage ipython magic,
    require memory_profiler package to be installed

    to get usage: %memit?

    Author: Vlad Niculae <vlad@vene.ro>
    Makes use of memory_profiler from Fabian Pedregosa
    available at https://github.com/fabianp/memory_profiler
    """
    from IPython.core.magic import Magics, line_magic, magics_class
    from memory_profiler import memory_usage as _mu

    try:
        ip = get_ipython()
    except NameError as err:
        raise Exception('not in ipython/jupyter kernel:\n {}'.format(err))

    @magics_class
    class MemMagics(Magics):
        @line_magic
        def memit(self, line='', setup='pass'):
            """Measure memory usage of a Python statement

            Usage, in line mode:
              %memit [-ir<R>t<T>] statement

            Options:
            -r<R>: repeat the loop iteration <R> times and take the best result.
            Default: 3

            -i: run the code in the current environment, without forking a new
            process. This is required on some MacOS versions of Accelerate if your
            line contains a call to `np.dot`.

            -t<T>: timeout after <T> seconds. Unused if `-i` is active.
            Default: None

            Examples
            --------
            ::

              In [1]: import numpy as np

              In [2]: %memit np.zeros(1e7)
              maximum of 3: 76.402344 MB per loop

              In [3]: %memit np.ones(1e6)
              maximum of 3: 7.820312 MB per loop

              In [4]: %memit -r 10 np.empty(1e8)
              maximum of 10: 0.101562 MB per loop

              In [5]: memit -t 3 while True: pass;
              Subprocess timed out.
              Subprocess timed out.
              Subprocess timed out.
              ERROR: all subprocesses exited unsuccessfully. Try again with the
              `-i` option.
              maximum of 3: -inf MB per loop

            """
            opts, stmt = self.parse_options(line, 'r:t:i', posix=False,
                                            strict=False)
            repeat = int(getattr(opts, 'r', 3))
            if repeat < 1:
                repeat == 1
            timeout = int(getattr(opts, 't', 0))
            if timeout <= 0:
                timeout = None
            run_in_place = hasattr(opts, 'i')

            # Don't depend on multiprocessing:
            try:
                import multiprocessing as pr
                from multiprocessing.queues import SimpleQueue
                q = SimpleQueue()
            except ImportError:
                class ListWithPut(list):
                    "Just a list where the `append` method is aliased to `put`."

                    def put(self, x):
                        self.append(x)

                q = ListWithPut()
                print('WARNING: cannot import module `multiprocessing`. Forcing '
                      'the `-i` option.')
                run_in_place = True

            ns = self.shell.user_ns

            def _get_usage(q, stmt, setup='pass', ns={}):
                try:
                    exec(setup) in ns
                    _mu0 = _mu()[0]
                    exec(stmt) in ns
                    _mu1 = _mu()[0]
                    q.put(_mu1 - _mu0)
                except Exception as e:
                    q.put(float('-inf'))
                    raise e

            if run_in_place:
                for _ in xrange(repeat):
                    _get_usage(q, stmt, ns=ns)
            else:
                # run in consecutive subprocesses
                at_least_one_worked = False
                for _ in xrange(repeat):
                    p = pr.Process(target=_get_usage, args=(q, stmt, 'pass', ns))
                    p.start()
                    p.join(timeout=timeout)
                    if p.exitcode == 0:
                        at_least_one_worked = True
                    else:
                        p.terminate()
                        if p.exitcode == None:
                            print('Subprocess timed out.')
                        else:
                            print('Subprocess exited with code %d.' % p.exitcode)
                        q.put(float('-inf'))

                if not at_least_one_worked:
                    print('ERROR: all subprocesses exited unsuccessfully. Try '
                          'again with the `-i` option.')

            usages = [q.get() for _ in xrange(repeat)]
            usage = max(usages)
            print("maximum of %d: %f MB per loop" % (repeat, usage))

    ip.register_magics(MemMagics)


#!/usr/bin/env python
# -- coding: utf-8 --
""" a package to manipulate the physical units of JSON data,
via unitschema

jsonextended utilises the pint package to manage physical units.
This was chosen due to its large user-base and support for numpy:
https://socialcompare.com/en/comparison/python-units-quantities-packages-383avix4

"""

from jsonextended.units.core import (get_in_units, apply_unitschema,
                                     split_quantities, combine_quantities)



#!/usr/bin/env python


import glob
import imp
import inspect
import os
import uuid
import warnings
from fnmatch import fnmatch
from contextlib import contextmanager

# py 2/3 compatibility
try:
    basestring
except NameError:
    basestring = str
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
try:
    from importlib.machinery import SourceFileLoader
    from types import ModuleType

    def load_source(modname, fname):
        loader = SourceFileLoader(modname, fname)
        mod = ModuleType(loader.name)
        loader.exec_module(mod)
        return mod

except ImportError as err:
    load_source = lambda modname, fname: imp.load_source(modname, fname)

from jsonextended.utils import get_module_path

# list of plugin categories,
# and their minimal class attribute interface
# must include plugin_name, plugin_descript
_plugins_interface = {
    'encoders': ['plugin_name', 'plugin_descript', 'objclass'],
    'decoders': ['plugin_name', 'plugin_descript', 'dict_signature'],
    'parsers': ['plugin_name', 'plugin_descript', 'file_regex', 'read_file']}

# builtin plugin locations
from jsonextended import encoders, parsers

_plugins_builtin = {'encoders': get_module_path(encoders),
                    'decoders': get_module_path(encoders),
                    'parsers': get_module_path(parsers)}

# the internal plugin store
_all_plugins = {name: {} for name in _plugins_interface}


def view_interfaces(category=None):
    """ return a view of the plugin minimal class attribute interface(s)

    Parameters
    ----------
    category : None or str
        if str, apply for single plugin category

    Examples
    --------

    >>> from pprint import pprint
    >>> pprint(view_interfaces())
    {'decoders': ['plugin_name', 'plugin_descript', 'dict_signature'],
     'encoders': ['plugin_name', 'plugin_descript', 'objclass'],
     'parsers': ['plugin_name', 'plugin_descript', 'file_regex', 'read_file']}

    """
    if category is not None:
        return sorted(_plugins_interface[category][:])
    else:
        return {k: v[:] for k, v in _plugins_interface.items()}


def view_plugins(category=None):
    """ return a view of the loaded plugin names and descriptions

    Parameters
    ----------
    category : None or str
        if str, apply for single plugin category

    Examples
    --------

    >>> from pprint import pprint
    >>> pprint(view_plugins())
    {'decoders': {}, 'encoders': {}, 'parsers': {}}

    >>> class DecoderPlugin(object):
    ...     plugin_name = 'example'
    ...     plugin_descript = 'a decoder for dicts containing _example_ key'
    ...     dict_signature = ('_example_',)
    ...
    >>> errors = load_plugin_classes([DecoderPlugin])

    >>> pprint(view_plugins())
    {'decoders': {'example': 'a decoder for dicts containing _example_ key'},
     'encoders': {},
     'parsers': {}}

    >>> view_plugins('decoders')
    {'example': 'a decoder for dicts containing _example_ key'}

    >>> unload_all_plugins()

    """
    dct = _all_plugins
    if not category is None:
        if category == 'parsers':
            return {name: {"descript": klass.plugin_descript, "regex": klass.file_regex}
                    for name, klass in _all_plugins[category].items()}
        return {name: klass.plugin_descript for name, klass in _all_plugins[category].items()}
    else:
        return {cat: {name: klass.plugin_descript
                      for name, klass in plugins.items()} for cat, plugins in _all_plugins.items()}


def get_plugins(category):
    """ get plugins for category """
    return _all_plugins[category]


def unload_all_plugins(category=None):
    """ clear all plugins

    Parameters
    ----------
    category : None or str
        if str, apply for single plugin category

    Examples
    --------

    >>> from pprint import pprint
    >>> pprint(view_plugins())
    {'decoders': {}, 'encoders': {}, 'parsers': {}}

    >>> class DecoderPlugin(object):
    ...     plugin_name = 'example'
    ...     plugin_descript = 'a decoder for dicts containing _example_ key'
    ...     dict_signature = ('_example_',)
    ...
    >>> errors = load_plugin_classes([DecoderPlugin])

    >>> pprint(view_plugins())
    {'decoders': {'example': 'a decoder for dicts containing _example_ key'},
     'encoders': {},
     'parsers': {}}

    >>> unload_all_plugins()
    >>> pprint(view_plugins())
    {'decoders': {}, 'encoders': {}, 'parsers': {}}

    """
    if category is None:
        for cat in _all_plugins:
            _all_plugins[cat] = {}
    else:
        _all_plugins[category] = {}


def unload_plugin(name, category=None):
    """ remove single plugin

    Parameters
    ----------
    name : str
        plugin name
    category : str
        plugin category

    Examples
    --------

    >>> from pprint import pprint
    >>> pprint(view_plugins())
    {'decoders': {}, 'encoders': {}, 'parsers': {}}

    >>> class DecoderPlugin(object):
    ...     plugin_name = 'example'
    ...     plugin_descript = 'a decoder for dicts containing _example_ key'
    ...     dict_signature = ('_example_',)
    ...
    >>> errors = load_plugin_classes([DecoderPlugin],category='decoders')

    >>> pprint(view_plugins())
    {'decoders': {'example': 'a decoder for dicts containing _example_ key'},
     'encoders': {},
     'parsers': {}}

    >>> unload_plugin('example','decoders')
    >>> pprint(view_plugins())
    {'decoders': {}, 'encoders': {}, 'parsers': {}}

    """
    if category is not None:
        _all_plugins[category].pop(name)
    else:
        for cat in _all_plugins:
            if name in _all_plugins[cat]:
                _all_plugins[cat].pop(name)


def load_plugin_classes(classes, category=None, overwrite=False):
    """ load plugins from class objects

    Parameters
    ----------
    classes: list
        list of classes
    category : None or str
        if str, apply for single plugin category
    overwrite : bool
        if True, allow existing plugins to be overwritten

    Examples
    --------

    >>> from pprint import pprint
    >>> pprint(view_plugins())
    {'decoders': {}, 'encoders': {}, 'parsers': {}}

    >>> class DecoderPlugin(object):
    ...     plugin_name = 'example'
    ...     plugin_descript = 'a decoder for dicts containing _example_ key'
    ...     dict_signature = ('_example_',)
    ...
    >>> errors = load_plugin_classes([DecoderPlugin])

    >>> pprint(view_plugins())
    {'decoders': {'example': 'a decoder for dicts containing _example_ key'},
     'encoders': {},
     'parsers': {}}

    >>> unload_all_plugins()

    """
    load_errors = []
    for klass in classes:
        for pcat, pinterface in _plugins_interface.items():
            if category is not None and not pcat == category:
                continue
            if all([hasattr(klass, attr) for attr in pinterface]):
                if klass.plugin_name in _all_plugins[pcat] and not overwrite:
                    err = '{0} is already set for {1}'.format(klass.plugin_name, pcat)
                    load_errors.append((klass.__name__, '{}'.format(err)))
                    continue
                _all_plugins[pcat][klass.plugin_name] = klass()
            else:
                load_errors.append((klass.__name__, 'does not match {0} interface: {1}'.format(pcat, pinterface)))
    return load_errors


@contextmanager
def plugins_context(classes, category=None):
    """ context manager to load plugin class(es) then unload on exit

    Parameters
    ----------
    classes: list
        list of classes
    category : None or str
        if str, apply for single plugin category

    Examples
    --------
    >>> from pprint import pprint

    >>> class DecoderPlugin(object):
    ...     plugin_name = 'example'
    ...     plugin_descript = 'a decoder for dicts containing _example_ key'
    ...     dict_signature = ('_example_',)
    ...

    >>> with plugins_context([DecoderPlugin]):
    ...     pprint(view_plugins())
    {'decoders': {'example': 'a decoder for dicts containing _example_ key'},
     'encoders': {},
     'parsers': {}}

    >>> pprint(view_plugins())
    {'decoders': {}, 'encoders': {}, 'parsers': {}}


    """
    original = {cat: list(_all_plugins[cat].keys()) for cat in _all_plugins}
    errors = load_plugin_classes(classes, category, overwrite=True)
    # if errors:
    #     for cat in _all_plugins:
    #         for name, kls in list(_all_plugins[cat].items()):
    #             if name not in original[cat]:
    #                 _all_plugins[cat].pop(name)
    #     raise RuntimeError("errors occurred while loading plugins: {}".format(errors))
    yield
    for cat in _all_plugins:
        for name, kls in list(_all_plugins[cat].items()):
            if name not in original[cat]:
                _all_plugins[cat].pop(name)


def load_plugins_dir(path, category=None, overwrite=False):
    """ load plugins from a directory

    Parameters
    ----------
    path : str or path-like
    category : None or str
        if str, apply for single plugin category
    overwrite : bool
        if True, allow existing plugins to be overwritten

    """
    # get potential plugin python files
    if hasattr(path, 'glob'):
        pypaths = path.glob('*.py')
    else:
        pypaths = glob.glob(os.path.join(path, '*.py'))

    load_errors = []
    for pypath in pypaths:
        # use uuid to ensure no conflicts in name space
        mod_name = str(uuid.uuid4())
        try:
            if hasattr(pypath, 'resolve'):
                # Make the path absolute, resolving any symlinks
                pypath = pypath.resolve()

            with warnings.catch_warnings(record=True):
                warnings.filterwarnings("ignore", category=ImportWarning)

                # for MockPaths
                if hasattr(pypath, 'maketemp'):
                    with pypath.maketemp() as f:
                        module = load_source(mod_name, f.name)
                else:
                    module = load_source(mod_name, str(pypath))

        except Exception as err:
            load_errors.append((str(pypath), 'Load Error: {}'.format(err)))
            continue

        # only get classes that are local to the module
        classes = [klass for klass_name, klass in inspect.getmembers(module, inspect.isclass) if
                   klass.__module__ == mod_name]
        load_errors += load_plugin_classes(classes, category, overwrite)

    return load_errors


def load_builtin_plugins(category=None, overwrite=False):
    """load plugins from builtin directories

    Parameters
    ----------
    name: None or str
    category : None or str
        if str, apply for single plugin category

    Examples
    --------

    >>> from pprint import pprint
    >>> pprint(view_plugins())
    {'decoders': {}, 'encoders': {}, 'parsers': {}}

    >>> errors = load_builtin_plugins()
    >>> errors
    []

    >>> pprint(view_plugins(),width=200)
    {'decoders': {'decimal.Decimal': 'encode/decode Decimal type',
                  'fractions.Fraction': 'encode/decode Fraction type',
                  'numpy.ndarray': 'encode/decode numpy.ndarray',
                  'pint.Quantity': 'encode/decode pint.Quantity object',
                  'python.set': 'decode/encode python set'},
     'encoders': {'decimal.Decimal': 'encode/decode Decimal type',
                  'fractions.Fraction': 'encode/decode Fraction type',
                  'numpy.ndarray': 'encode/decode numpy.ndarray',
                  'pint.Quantity': 'encode/decode pint.Quantity object',
                  'python.set': 'decode/encode python set'},
     'parsers': {'csv.basic': 'read *.csv delimited file with headers to {header:[column_values]}',
                 'csv.literal': 'read *.literal.csv delimited files with headers to {header:column_values}, with number strings converted to int/float',
                 'hdf5.read': 'read *.hdf5 (in read mode) files using h5py',
                 'ipynb': 'read Jupyter Notebooks',
                 'json.basic': 'read *.json files using json.load',
                 'keypair': "read *.keypair, where each line should be; '<key> <pair>'",
                 'yaml.ruamel': 'read *.yaml files using ruamel.yaml'}}


    >>> unload_all_plugins()

    """
    load_errors = []
    for cat, path in _plugins_builtin.items():
        if cat != category and not category is None:
            continue
        load_errors += load_plugins_dir(path, cat, overwrite=overwrite)
    return load_errors


def encode(obj, outtype='json', raise_error=False):
    """ encode objects, via encoder plugins, to new types

    Parameters
    ----------
    outtype: str
        use encoder method to_<outtype> to encode
    raise_error : bool
        if True, raise ValueError if no suitable plugin found

    Examples
    --------
    >>> load_builtin_plugins('encoders')
    []

    >>> from decimal import Decimal
    >>> encode(Decimal('1.3425345'))
    {'_python_Decimal_': '1.3425345'}
    >>> encode(Decimal('1.3425345'),outtype='str')
    '1.3425345'

    >>> encode(set([1,2,3,4,4]))
    {'_python_set_': [1, 2, 3, 4]}
    >>> encode(set([1,2,3,4,4]),outtype='str')
    '{1, 2, 3, 4}'

    >>> unload_all_plugins()

    """
    for encoder in get_plugins('encoders').values():
        if isinstance(obj, encoder.objclass) and hasattr(encoder, 'to_{}'.format(outtype)):
            return getattr(encoder, 'to_{}'.format(outtype))(obj)
            break

    if raise_error:
        raise ValueError('No JSON serializer is available for {0} (of type {1})'.format(obj, type(obj)))
    else:
        return obj


def decode(dct, intype='json', raise_error=False):
    """ decode dict objects, via decoder plugins, to new type

    Parameters
    ----------
    intype: str
        use decoder method from_<intype> to encode
    raise_error : bool
        if True, raise ValueError if no suitable plugin found

    Examples
    --------
    >>> load_builtin_plugins('decoders')
    []

    >>> from decimal import Decimal
    >>> decode({'_python_Decimal_':'1.3425345'})
    Decimal('1.3425345')

    >>> unload_all_plugins()

    """
    for decoder in get_plugins('decoders').values():
        if set(list(decoder.dict_signature)).issubset(dct.keys()) and hasattr(decoder, 'from_{}'.format(intype)) \
                and getattr(decoder, 'allow_other_keys', False):
            return getattr(decoder, 'from_{}'.format(intype))(dct)
            break
        elif sorted(list(decoder.dict_signature)) == sorted(dct.keys()) and hasattr(decoder, 'from_{}'.format(intype)):
            return getattr(decoder, 'from_{}'.format(intype))(dct)
            break

    if raise_error:
        raise ValueError('no suitable plugin found for: {}'.format(dct))
    else:
        return dct


def parser_available(fpath):
    """ test if parser plugin available for fpath

    Examples
    --------

    >>> load_builtin_plugins('parsers')
    []
    >>> test_file = StringIO('{"a":[1,2,3.4]}')
    >>> test_file.name = 'test.json'
    >>> parser_available(test_file)
    True
    >>> test_file.name = 'test.other'
    >>> parser_available(test_file)
    False

    >>> unload_all_plugins()

    """
    if isinstance(fpath, basestring):
        fname = fpath
    elif hasattr(fpath, 'open') and hasattr(fpath, 'name'):
        fname = fpath.name
    elif hasattr(fpath, 'readline') and hasattr(fpath, 'name'):
        fname = fpath.name
    else:
        raise ValueError('fpath should be a str or file_like object: {}'.format(fpath))

    for parser in get_plugins('parsers').values():
        if fnmatch(fname, parser.file_regex):
            return True

    return False


def parse(fpath, **kwargs):
    """ parse file contents, via parser plugins, to dict like object
    NB: the longest file regex will be used from plugins

    Parameters
    ----------
    fpath : file_like
         string, object with 'open' and 'name' attributes, or
         object with 'readline' and 'name' attributes
    kwargs : keywords
        to pass to parser plugin

    Examples
    --------

    >>> load_builtin_plugins('parsers')
    []

    >>> from pprint import pformat

    >>> json_file = StringIO('{"a":[1,2,3.4]}')
    >>> json_file.name = 'test.json'

    >>> dct = parse(json_file)
    >>> print(pformat(dct).replace("u'","'"))
    {'a': [1, 2, 3.4]}

    >>> reset = json_file.seek(0)
    >>> from decimal import Decimal
    >>> dct = parse(json_file, parse_float=Decimal,other=1)
    >>> print(pformat(dct).replace("u'","'"))
    {'a': [1, 2, Decimal('3.4')]}

    >>> class NewParser(object):
    ...     plugin_name = 'example'
    ...     plugin_descript = 'loads test.json files'
    ...     file_regex = 'test.json'
    ...     def read_file(self, file_obj, **kwargs):
    ...         return {'example':1}
    >>> load_plugin_classes([NewParser],'parsers')
    []
    >>> reset = json_file.seek(0)
    >>> parse(json_file)
    {'example': 1}

    >>> unload_all_plugins()

    """
    if isinstance(fpath, basestring):
        fname = fpath
    elif hasattr(fpath, 'open') and hasattr(fpath, 'name'):
        fname = fpath.name
    elif hasattr(fpath, 'readline') and hasattr(fpath, 'name'):
        fname = fpath.name
    else:
        raise ValueError('fpath should be a str or file_like object: {}'.format(fpath))

    parser_dict = {plugin.file_regex: plugin for plugin in get_plugins('parsers').values()}

    # find longest match first
    for regex in sorted(parser_dict.keys(), key=len, reverse=True):
        parser = parser_dict[regex]
        if fnmatch(fname, regex):
            if isinstance(fpath, basestring):
                with open(fpath, 'r') as file_obj:
                    data = parser.read_file(file_obj, **kwargs)
            elif hasattr(fpath, 'open'):
                with fpath.open('r') as file_obj:
                    data = parser.read_file(file_obj, **kwargs)
            elif hasattr(fpath, 'readline'):
                data = parser.read_file(fpath, **kwargs)
            return data

    raise ValueError('{} does not match any regex'.format(fname))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Setup for jsonextended."""

import io
from importlib import import_module
from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = f.read().splitlines()
with open('test_requirements.txt') as f:
    test_requirements = f.read().splitlines()

with io.open('README.rst') as readme:
    setup(
        name='jsonextended',
        version=import_module('jsonextended').__version__,
        description='Extending the python json package functionality',
        long_description=readme.read(),
        install_requires=requirements,
        tests_require=test_requirements,
        license='MIT',
        author='Chris Sewell',
        author_email='chrisj_sewell@hotmail.com',
        url='https://github.com/chrisjsewell/jsonextended',
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Environment :: Console',
            'Environment :: Web Environment',
            'Intended Audience :: End Users/Desktop',
            'Intended Audience :: Science/Research',
            'Intended Audience :: Financial and Insurance Industry',
            'License :: OSI Approved :: MIT License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.2',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Topic :: Scientific/Engineering :: Chemistry',
            'Topic :: Scientific/Engineering :: Physics',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Topic :: Utilities',
        ],
        keywords='json, units, parser, python',
        zip_safe=True,
        packages = find_packages(),
        package_data={'': ['*.json', '*.crystal.out','*.csv','*.keypair','*.data']},
    )



#!/usr/bin/env python
import json


class NBParser(object):
    """
    Examples
    --------
    >>> from jsonextended.utils import MockPath
    >>> from jsonextended.edict import pprint
    >>> fileobj = MockPath(is_file=True,
    ... content='''{
    ... "cells":[],
    ... "metadata":{}
    ... }'''
    ... )
    >>> with fileobj.open() as f:
    ...     data = NBParser().read_file(f)
    >>> pprint(data)
    cells: []
    metadata:

    """

    plugin_name = 'ipynb'
    plugin_descript = 'read Jupyter Notebooks'
    file_regex = '*.ipynb'

    def read_file(self, file_obj, **kwargs):
        return json.load(file_obj)



#!/usr/bin/env python

from ruamel.yaml import YAML


class YAML_Parser(object):
    r"""

    Examples
    --------

    >>> from pprint import pprint
    >>> from jsonextended.utils import MockPath
    >>> fileobj = MockPath(is_file=True,
    ... content='key1:\n  subkey1: a\n  subkey2: [1,2,3]'
    ... )
    >>> with fileobj.open() as f:
    ...     data = YAML_Parser().read_file(f)
    >>> pprint(dict(data["key1"]))
    {'subkey1': 'a', 'subkey2': [1, 2, 3]}

    """

    plugin_name = 'yaml.ruamel'
    plugin_descript = 'read *.yaml files using ruamel.yaml'
    file_regex = '*.yaml'

    def read_file(self, file_obj, **kwargs):
        return YAML().load(file_obj)



#!/usr/bin/env python
# -- coding: utf-8 --

from fnmatch import fnmatch

# make units optional when importing jsonextended
try:
    import numpy as np
except ImportError:
    pass

try:
    from pint import UnitRegistry

    ureg = UnitRegistry()
    from pint.quantity import _Quantity
except ImportError:
    pass

from jsonextended.edict import flatten, flatten2d, unflatten, merge


def get_in_units(value, units):
    """get a value in the required units """
    try:
        return ureg.Quantity(value, units)
    except NameError:
        raise ImportError('please install pint to use this module')


def apply_unitschema(data, uschema, as_quantity=True,
                     raise_outerr=False, convert_base=False,
                     use_wildcards=False, list_of_dicts=False):
    """ apply the unit schema to the data

    Parameters
    ----------
    data : dict
    uschema : dict
        units schema to apply
    as_quantity : bool
        if true, return values as pint.Quantity objects
    raise_outerr : bool
        raise error if a unit cannot be found in the outschema
    convert_to_base : bool
        rescale units to base units
    use_wildcards : bool
        if true, can use * (matches everything) and ? (matches any single character)
    list_of_dicts: bool
        treat list of dicts as additional branches

    Examples
    --------
    >>> from pprint import pprint

    >>> data = {'energy':1,'x':[1,2],'other':{'y':[4,5]},'y':[4,5],'meta':None}
    >>> uschema =   {'energy':'eV','x':'nm','other':{'y':'m'},'y':'cm'}
    >>> data_units = apply_unitschema(data,uschema)
    >>> pprint(data_units)
    {'energy': <Quantity(1, 'electron_volt')>,
     'meta': None,
     'other': {'y': <Quantity([4 5], 'meter')>},
     'x': <Quantity([1 2], 'nanometer')>,
     'y': <Quantity([4 5], 'centimeter')>}

    >>> newschema = {'energy':'kJ','other':{'y':'nm'},'y':'m'}
    >>> new_data = apply_unitschema(data_units,newschema)
    >>> str(new_data["energy"])
    '1.60217653e-22 kilojoule'
    >>> new_data["other"]["y"].magnitude.round(3).tolist(), str(new_data["other"]["y"].units)
    ([4000000000.0, 5000000000.0], 'nanometer')

    >>> old_data = apply_unitschema(new_data,uschema,as_quantity=False)
    >>> old_data["energy"]
    1.0
    >>> old_data["other"]["y"].round(3).tolist()
    [4.0, 5.0]

    """
    try:
        _Quantity
    except NameError:
        raise ImportError('please install pint to use this module')
    list_of_dicts = '__list__' if list_of_dicts else None

    # flatten edict
    uschema_flat = flatten(uschema, key_as_tuple=True)
    # sorted by longest key size, to get best match first
    uschema_keys = sorted(uschema_flat, key=len, reverse=True)
    data_flat = flatten(data, key_as_tuple=True, list_of_dicts=list_of_dicts)

    for dkey, dvalue in data_flat.items():
        converted = False
        for ukey in uschema_keys:

            if not len(ukey) == len(dkey[-len(ukey):]):
                continue

            if use_wildcards:
                match = all([fnmatch(d, u) for u, d in zip(ukey, dkey[-len(ukey):])])
            else:
                match = ukey == dkey[-len(ukey):]

            if match:
                # handle the fact that it return an numpy object type if list of floats
                if isinstance(dvalue, (list, tuple)):
                    dvalue = np.array(dvalue)
                    if dvalue.dtype == np.object:
                        dvalue = dvalue.astype(float)

                if isinstance(dvalue, _Quantity):
                    quantity = dvalue.to(uschema_flat[ukey])
                else:
                    quantity = ureg.Quantity(dvalue, uschema_flat[ukey])

                if convert_base:
                    quantity = quantity.to_base_units()

                if as_quantity:
                    data_flat[dkey] = quantity
                else:
                    data_flat[dkey] = quantity.magnitude
                break

        if not converted and raise_outerr:
            raise KeyError('could not find units for {}'.format(dkey))

    return unflatten(data_flat, list_of_dicts=list_of_dicts)


def split_quantities(data, units='units', magnitude='magnitude',
                     list_of_dicts=False):
    """ split pint.Quantity objects into <unit,magnitude> pairs

    Parameters
    ----------
    data : dict
    units : str
        name for units key
    magnitude : str
        name for magnitude key
    list_of_dicts: bool
        treat list of dicts as additional branches

    Examples
    --------
    >>> from pprint import pprint

    >>> from pint import UnitRegistry
    >>> ureg = UnitRegistry()
    >>> Q = ureg.Quantity

    >>> qdata = {'energy': Q(1.602e-22, 'kilojoule'),
    ...          'meta': None,
    ...          'other': {'y': Q([4,5,6], 'nanometer')},
    ...          'x': Q([1,2,3], 'nanometer'),
    ...          'y': Q([8,9,10], 'meter')}
    ...
    >>> split_data = split_quantities(qdata)
    >>> pprint(split_data)
    {'energy': {'magnitude': 1.602e-22, 'units': 'kilojoule'},
     'meta': None,
     'other': {'y': {'magnitude': array([4, 5, 6]), 'units': 'nanometer'}},
     'x': {'magnitude': array([1, 2, 3]), 'units': 'nanometer'},
     'y': {'magnitude': array([ 8,  9, 10]), 'units': 'meter'}}

    """
    try:
        _Quantity
    except NameError:
        raise ImportError('please install pint to use this module')
    list_of_dicts = '__list__' if list_of_dicts else None
    data_flatten = flatten(data, list_of_dicts=list_of_dicts)
    for key, val in data_flatten.items():
        if isinstance(val, _Quantity):
            data_flatten[key] = {units: str(val.units),
                                 magnitude: val.magnitude}
    return unflatten(data_flatten, list_of_dicts=list_of_dicts)


def combine_quantities(data, units='units', magnitude='magnitude',
                       list_of_dicts=False):
    """ combine <unit,magnitude> pairs into pint.Quantity objects

    Parameters
    ----------
    data : dict
    units : str
        name of units key
    magnitude : str
        name of magnitude key
    list_of_dicts: bool
        treat list of dicts as additional branches

    Examples
    --------
    >>> from pprint import pprint

    >>> sdata = {'energy': {'magnitude': 1.602e-22, 'units': 'kilojoule'},
    ...          'meta': None,
    ...          'other': {'y': {'magnitude': [4, 5, 6], 'units': 'nanometer'}},
    ...          'x': {'magnitude': [1, 2, 3], 'units': 'nanometer'},
    ...          'y': {'magnitude': [8,9,10], 'units': 'meter'}}
    ...
    >>> combined_data = combine_quantities(sdata)
    >>> pprint(combined_data)
    {'energy': <Quantity(1.602e-22, 'kilojoule')>,
     'meta': None,
     'other': {'y': <Quantity([4 5 6], 'nanometer')>},
     'x': <Quantity([1 2 3], 'nanometer')>,
     'y': <Quantity([ 8  9 10], 'meter')>}

    """
    try:
        _Quantity
    except NameError:
        raise ImportError('please install pint to use this module')
    list_of_dicts = '__list__' if list_of_dicts else None

    data_flatten2d = flatten2d(data, list_of_dicts=list_of_dicts)
    new_dict = {}
    for key, val in list(data_flatten2d.items()):
        if units in val and magnitude in val:
            quantity = ureg.Quantity(val.pop(magnitude), val.pop(units))
            if not val:
                data_flatten2d.pop(key)
            new_dict[key] = quantity
    final_dict = merge([data_flatten2d, new_dict])
    # olddict = unflatten(data_flatten2d,list_of_dicts=list_of_dicts)
    # new_dict = unflatten(new_dict,list_of_dicts=list_of_dicts)
    return unflatten(final_dict, list_of_dicts=list_of_dicts)  # merge([olddict,new_dict])


if __name__ == '__main__':
    import doctest

    print(doctest.testmod())
